package com.biganto.visual.roompark.presentation.screen.photo.util



import android.content.Context
import android.graphics.*
import android.graphics.drawable.Drawable
import android.view.Surface
import com.bumptech.glide.Glide
import com.bumptech.glide.load.engine.DiskCacheStrategy
import com.bumptech.glide.request.RequestOptions
import com.bumptech.glide.request.target.CustomTarget
import com.bumptech.glide.request.transition.Transition
import timber.log.Timber
import kotlin.math.abs
import kotlin.math.absoluteValue
import kotlin.math.sign
import kotlin.math.sin


/**
 * Created by Vladislav Bogdashkin on 18.07.2019.
 */

private const val CONST_SLEEP_THRESHOLD = 8L
private const val CONST_CENTER_PHOTO_BORDER_PX = 8
private const val CONST_CENTER_HEIGHT_DIFF_PX = 0
private const val CONST_HEIGHT_MARGIN_BORDER_PX = 10
private const val CONST_DIFF_BORDER_PX = 1
private const val CONST_DECAY_TIMER = 500L

class RendererThread(val context: Context, val surface: SurfaceTexture)
    : Thread(){

    private val dens:Float by lazy { context.resources.displayMetrics.density }

    private val SLEEP_THRESHOLD get() = CONST_SLEEP_THRESHOLD
    private val CENTER_PHOTO_BORDER_PX get() =
            (CONST_CENTER_PHOTO_BORDER_PX *dens).int
    private val CENTER_HEIGHT_DIFF_PX get() =
                (CONST_CENTER_HEIGHT_DIFF_PX *dens).int
    private val HEIGHT_MARGIN_BORDER_PX get() =
                    (CONST_HEIGHT_MARGIN_BORDER_PX *dens).int
    private val DIFF_BORDER_PX get() =
                        (CONST_DIFF_BORDER_PX *dens).int
    private val DECAY_TIMER get() = CONST_DECAY_TIMER

    private var isDirty = true
    var isStopped = false

    private var width = 0
    private var height = 0

    private var bmpCache:MutableList<Bitmap?> = mutableListOf()
    private var holdIndex = 0
        get() = if (isDragging) field
        else currentIndex

    private var currentIndex:Int = 0

    set(value) {Timber.w("currentIndex to: $value");field = value}
    private var bmpCacheUrls:List<String> = arrayListOf()

    fun setAdapter(urls:List<String> ) {
        bmpCacheUrls = urls
        bmpCache = MutableList(bmpCacheUrls.size) { null}
        bmpCacheUrls.fetchUrls()
        bordersArray = IntArray(bmpCacheUrls.size*4)
        cropArray = IntArray(bmpCacheUrls.size*4)

    }

    private fun List<String>.fetchUrls(){
        this.forEachIndexed{index, s ->
            if (bmpCache.getOrNull(index) == null){
                Timber.d("loading: $s")
                Glide.with(context)
                    .asBitmap()
                    .load(bmpCacheUrls[index])
                    .apply(RequestOptions().override(width/bmpCache.size, height))
                    .diskCacheStrategy(DiskCacheStrategy.ALL)
                    .into(object : CustomTarget<Bitmap>(){
                        override fun onLoadCleared(placeholder: Drawable?) {
                            Timber.d(" onLoadCleared ")
                            bmpCache[index] = null
                        }

                        override fun onResourceReady(
                            resource: Bitmap,
                            transition: Transition<in Bitmap>?
                        ) {
                            Timber.d(" onResourceReady $resource")
                            bmpCache[index] = resource
                        }
                    })
            }
        }
    }


    fun setIndex(index:Int){
        if (!isDragging) holdIndex = index
        currentIndex = index
        isDirty=true
    }
    private var bordersArray:IntArray? = null
    private var cropArray:IntArray? = null

    private fun IntArray.getPosition(dir:Float):Int {
        if (dir < 0 && this.getX(currentIndex) > width / 2f)
            return (currentIndex - 1).coerceIn(0,this.size/4-1)

        if (dir > 0 && this.getRight(currentIndex) < width / 2f)
            return (currentIndex + 1).coerceIn(0,this.size/4-1)

        return currentIndex
    }

    private fun IntArray.getX(index: Int) = if (index>=0) this[index*4+0] else 0
    private fun IntArray.getY(index: Int) = this[index*4+1]
    private fun IntArray.getRight(index: Int) = if (index>=0) this[index*4+2] else 0
    private fun IntArray.getBottom(index: Int) = this[index*4+3]

    private fun IntArray.getWidth(index: Int) = this[index*4+2] - this[index*4+0]
    private fun IntArray.getHeight(index: Int) = this[index*4+3] - this[index*4+1]
    private fun IntArray.aspect(index: Int) =
        this.getWidth(index).toFloat()/this.getHeight(index).toFloat()

    private fun IntArray.setX(index:Int,block:()->Int){
        this[index*4+0] = block.invoke()
    }
    private fun IntArray.setY(index:Int,block:()->Int){
        this[index*4+1] = block.invoke()
    }
    private fun IntArray.setRight(index:Int,block:()->Int){
        this[index*4+2] = block.invoke()
    }
    private fun IntArray.setBottom(index:Int,block:()->Int){
        this[index*4+3] = block.invoke()
    }

    private val Bitmap.aspect:Float get()  = this.width.toFloat()/this.height.toFloat()

    private fun calcBorders() {
        val centerX = width / 2f - scrollOffset
        val adjWidth = height * .6f
        bmpCache.forEachIndexed { index, bitmap ->
            val limit = (2 - (index - currentIndex).absoluteValue).coerceIn(0, 1)
            val direction = (index - currentIndex).sign//cos(((index-currentIndex).sign)*1.57f)
            val ch2 = limit * (
                    (slidePosition * direction).coerceIn(0f, 1f)
                            + (1f - direction.absoluteValue) * (1f - slidePosition.absoluteValue)
                    )

            bordersArray?.let { arr ->
                val wDiff = height * (bitmap?.aspect ?: 0f)
                val _width = adjWidth +
                        dragKoef*((wDiff - adjWidth) / 2f + CENTER_PHOTO_BORDER_PX) * ch2

                arr.setX(index) { arr.getRight(index - 1) }
                arr.setRight(index) {(arr.getX(index) + _width).int}
                arr.setY(index) {(HEIGHT_MARGIN_BORDER_PX + CENTER_HEIGHT_DIFF_PX * (1f - ch2)).int}
                arr.setBottom(index) {
                    height - (HEIGHT_MARGIN_BORDER_PX + CENTER_HEIGHT_DIFF_PX * (1f - ch2)).int
                }
            }

            cropArray?.let { arr ->
                val asp = bordersArray?.aspect(index) ?: 0f

                bitmap?.let {
                    if (asp >= it.aspect) {
                        arr.setX(index) { 0 }
                        arr.setRight(index) { it.width }

                        val hKef = it.aspect / asp
                        val marg = it.height * (1f - hKef) / 2f

                        arr.setY(index) { marg.int }
                        arr.setBottom(index) { it.height - marg.int }
                    } else {

                        arr.setY(index) { 0 }
                        arr.setBottom(index) { it.height }

                        val hKef = asp / it.aspect
                        val marg = it.width * (1f - hKef) / 2f

                        arr.setX(index) { marg.int }
                        arr.setRight(index) { it.width - marg.int }
                    }
                }
            }
        }

        bordersArray?.let { arr ->
            val halfW = arr.getWidth(currentIndex) / 2f
            val minusW = arr.getX(holdIndex) - centerX + halfW * (1f + 2f * slidePosition)
            bmpCache.forEachIndexed { index, bitmap ->
                val limit = (2 - (index - currentIndex).absoluteValue).coerceIn(0, 1)
                val direction = (index - currentIndex).sign
                val cBorder = (CENTER_PHOTO_BORDER_PX)
                val borderKoef = (
                        direction + 1f
                        - limit * sin((direction + slidePosition.sign).coerceIn(-1f, 1f) * 1.57f)
                        * slidePosition.absoluteValue
                        )

                val borderSize = dragKoef*cBorder * borderKoef

                arr.setX(index) {(arr.getX(index) - minusW.int + borderSize + DIFF_BORDER_PX).int}

                arr.setRight(index) {
                    (arr.getRight(index) - minusW.int + borderSize - DIFF_BORDER_PX).int
                }
            }
        }
    }

    private val Float.int:Int get() {return this.toInt()}

    fun perfomDrag(distanceX:Float?,dragging:Boolean = true){
        isDragging = dragging
        if (isDragging) distanceX?.let {perfomScroll(it)}
        else {
            holdIndex = currentIndex
            scrollOffset = 0f
            touchOffset=0f
        }
        isDirty = true
    }

    private fun perfomScroll(distanceX:Float) {
        touchOffset += distanceX * scrollScaleFactor
        val p = bordersArray?.getPosition(distanceX.sign)?:0
        if (p != currentIndex) {
            setIndex(p)
            onChangePage?.invoke(p)
        }
        isDirty=true
    }

    var onChangePage: ((newPage:Int)->Unit)? = null

    private var scrollScaleFactor = 1f
    private var touchOffset = 0f
    private var scrollOffset = 0f
    private var isDragging = false
    private var slidePosition = 0f
    private var dragKoef = 1f

    fun setSlidePosition(newPosition:Float){
        slidePosition = newPosition
        isDirty = true
    }

    private var decayFraction  = 0f

    private fun timerDecay(mills:Long){
        isDirty=true
        decayFraction+=mills
        animateBorders()
        scrollOffset = scrollOffset.smoothLerp(touchOffset-scrollOffset,.2f)
        if (touchOffset> 0 && scrollOffset >= touchOffset) {
            scrollOffset = touchOffset
            isDirty=false
        }
        else if (touchOffset< 0 && scrollOffset <= touchOffset) {
            scrollOffset = touchOffset
            isDirty=false
        }
        if (decayFraction> DECAY_TIMER) isDirty=false
        val dirty = animateBorders()
        if (!isDirty) isDirty = dirty
    }

    private fun animateBorders():Boolean{
        if (isDragging && dragKoef>0){
            dragKoef = dragKoef.smoothLerp(-dragKoef,.12f)
            if (dragKoef<.1f) dragKoef = 0f
            return  dragKoef>0
        }

        if (!isDragging && dragKoef<1f) {
            dragKoef = dragKoef.smoothLerp(1f-dragKoef,.03f)
            if (dragKoef>.99f) dragKoef = 1f
            return dragKoef<1f
        }
        return false
    }

    private fun Float.smoothLerp( delta: Float, fraction:Float =.4f): Float {
        return  this + fraction * (delta)
    }

    fun setSize(w: Int, h: Int){
        width=w
        height=h
    }

    override fun run() {
        super.run()
        Timber.w("isStopped: $isStopped")
        Timber.w("isDirty: $isDirty")
        isDirty = true

        while (!isStopped) {
            while (!isDirty)
                sleep(SLEEP_THRESHOLD) // in real life this sleep is more complicated
            isDirty = false

            val surf = Surface(surface)
            val dRect = Rect(
                width.div(4) + width.div(2)
                , HEIGHT_MARGIN_BORDER_PX
                , width - width.div(4)
                , height.minus(HEIGHT_MARGIN_BORDER_PX)
            )
            val c = surf.lockCanvas(dRect)
            c.drawColor(Color.TRANSPARENT)
            timerDecay(SLEEP_THRESHOLD)
            if (bmpCache.size>0) calcBorders()

            Timber.d("bmpCache ${bmpCache.size}")
            bmpCache.forEachIndexed { index, bitmap ->
                //                Timber.d("bitmap ${bitmap} / $index")
                bitmap?.let {

                    val bmpX = cropArray?.getX(index) ?: 0
                    val bmpY = cropArray?.getY(index) ?: 0
                    val bmpRight = cropArray?.getRight(index) ?: 0
                    val bmpBottom = cropArray?.getBottom(index) ?: 0


                    val dstX = bordersArray?.getX(index) ?: 0
                    val dstY = bordersArray?.getY(index) ?: 0
                    val dstRight = bordersArray?.getRight(index) ?: 0
                    val dstBottom = bordersArray?.getBottom(index) ?: 0

                    c.drawBitmap(
                        it
                        , Rect(bmpX, bmpY, bmpRight, bmpBottom)
                        , Rect(dstX, dstY, dstRight, dstBottom)
                        , null
                    )
                }
            }

            surf.unlockCanvasAndPost(c)
            surf.release()
//            isDirty = false
        }
        Timber.w("STOPPED; $isStopped")
        Timber.w("STOPPED; $isDirty")
        isDirty=false
    }

    private fun calcHeightPixel(index:Int, selectedIndex:Int): Int {
        return height
    }

    private fun calcWidthPixel(index:Int, selectedIndex:Int): Int {
        val dist = abs(index-selectedIndex)
        val bled = index - selectedIndex
//        Timber.d("gonna calc: $bled")
        try {

            var bled_res = 100;
//            Timber.d("res is : $bled_res")

            if (bled==0) bled_res = calcHeightPixel(index,selectedIndex)
            else bled_res = (calcHeightPixel(index,selectedIndex)* (1f - .05f*dist)).int

        bled_res = when(bled)
        {
             0 -> calcHeightPixel(index,selectedIndex)
            in -1 downTo Int.MIN_VALUE -> (calcHeightPixel(index,selectedIndex)* (1f - .05f*dist)).int
            in 1 .. Int.MAX_VALUE -> (calcHeightPixel(index,selectedIndex)* (1f - .05f*dist)).int
            else -> error("Unexeptable value")

        }
//            Timber.d("res is : $bled_res")
            return  bled_res
        }catch (e:Exception){Timber.e(e)}
        return  -1
    }

    private fun Bitmap.flip():Bitmap{
        val flip = Matrix()
        flip.postScale(1f, -1f)
        return Bitmap.createBitmap(
            this,
            0,
            0,
            this.width,
            this.height,
            flip,
            true
        )
    }

}

